/* Copyright (C) 1998 Alexander Kourakos */
/* Copyright (C) 2021 Arktis Desktop Project, GNU GPLv3 or any later  */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <X11/Xatom.h>
#include "SDClock.h"
#include "clock_bg.xpm"

static Display *display;
static Window root, window, main_window;
static GC gc;
static XFontSet fontset;
static Atom wm_delete_window, n_wm_1, n_wm_2, does, what[4];
static Pixel back_pixel, hand_pixel, second_hand_pixel, date_pixel;
static Pixmap back_pm, mask_pm, all_pm;
static Pixel GetColor(char *);
static int old_hour = 0, old_minute = 0, old_second = 0;
static int org_x, org_y, pos_x, pos_y, ix, iy;

int get_x(void) {
	FILE *fp;
	int x = option.xx;
	if ((fp = fopen("/root/.SDClock/conf/dx", "r")) != NULL) {
		if (fscanf(fp, "%d\n", &x) == -1)
		fprintf(stderr, "SDClock: x conf not set.. \n");
	fclose(fp);
	}
	return x;
}

int get_y(void) {
	FILE *fp;
	int y = option.yy;
	if ((fp = fopen("/root/.SDClock/conf/dy", "r")) != NULL) {
		if (fscanf(fp, "%d\n", &y) == -1)
		fprintf(stderr, "SDClock: y conf not set..  \n");
	fclose(fp);
	}
	return y;
}

void
CreateWindow(int ac, char *av[]) {
	int zero = 0;
	char *app_name = NAME;
	char *app_class = CLASS;
	int screen;
	XSizeHints size_hints;
	XClassHint class_hints;
	XTextProperty window_name;
	XGCValues gcv;
	unsigned long gcm;
	int result, use_internal_pixmap;
	display = XOpenDisplay(NULL);
	if (display == NULL) {
		fprintf(stderr, "SDClock: could not open display ``%s''..exit.\n",
				XDisplayName(NULL));
		exit(EXIT_FAILURE);
	}
	wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", FALSE);
	screen = DefaultScreen(display);
	root = RootWindow(display, screen);
	back_pixel = GetColor("black");
	hand_pixel = GetColor(option.hc);
	second_hand_pixel = GetColor(option.sc);
	date_pixel = GetColor(option.dc);
	use_internal_pixmap = TRUE;
	XpmAttributes attributes;
	attributes.valuemask = XpmSize;
	if (option.bg_pix[0] != '\0') {
		char db[88];
		sprintf(db, "/root/.SDClock/data/%s", option.bg_pix);
		result = XpmReadFileToPixmap(display, root, db, &back_pm, &mask_pm, 
										&attributes);
		if (result != XpmSuccess)
			fprintf(stderr, "SDClock: theme pixmap error.. \n");      
		else
			use_internal_pixmap = FALSE;
	}
	if (use_internal_pixmap) {
		result = XpmCreatePixmapFromData(display, root, clock_bg_xpm,
											&back_pm, &mask_pm, &attributes);
		if (result != XpmSuccess) {
			if (result == XpmColorFailed)
				fprintf(stderr, "SDClock: bad pixmap color.. exit.\n");
			else
				fprintf(stderr, "SDClock: built-in pixmap error.. exit.\n");
			exit(EXIT_FAILURE);
		}
	}
	ix = attributes.width;
	iy = attributes.height;
	size_hints.flags = PMinSize | PMaxSize | PPosition;
	size_hints.min_width = size_hints.max_width = ix;
	size_hints.min_height = size_hints.max_height = iy;
	XpmFreeAttributes(&attributes);
	all_pm = XCreatePixmap(display, root, ix, iy, 
							DefaultDepth(display, screen));
	XWMGeometry(display, screen, "", NULL, 0, &size_hints, &zero, &zero,
				&size_hints.width, &size_hints.height, &zero);
	window = XCreateSimpleWindow(display, root, get_x(), get_y(), 
			size_hints.width, size_hints.height, 0, back_pixel, back_pixel);
	main_window = window;
	XSetWMNormalHints(display, window, &size_hints);
	class_hints.res_name = app_name;
	class_hints.res_class = app_class;
	XSetClassHint(display, window, &class_hints);
	n_wm_1 = XInternAtom(display, "_NET_WM_WINDOW_TYPE", False);
	n_wm_2 = XInternAtom(display, "_NET_WM_WINDOW_TYPE_SPLASH", False);
	XChangeProperty(display, window, n_wm_1, XA_ATOM, 32, PropModeReplace,
					(unsigned char *) &n_wm_2, 1);
	does = XInternAtom(display, "_NET_WM_STATE", False);
	what[0] = XInternAtom(display, "_NET_WM_STATE_SKIP_PAGER", False);
	what[1] = XInternAtom(display, "_NET_WM_STATE_SKIP_TASKBAR", False);
	what[2] = XInternAtom(display, "_NET_WM_STATE_STICKY", False);
	what[3] = XInternAtom(display, "_NET_WM_STATE_BELOW", False);
	XChangeProperty(display, window, does, XA_ATOM, 32, PropModeAppend, 
					(unsigned char *) &what, 4);
	XSelectInput(display, main_window, StructureNotifyMask | ButtonPressMask | 
					Button1MotionMask | ButtonReleaseMask);
	if (XStringListToTextProperty(&app_name, 1, &window_name) == 0) {
		fprintf(stderr, "SDClock: can't allocate window name..exit.\n");
		exit(EXIT_FAILURE);
	}
	XSetWMName(display, window, &window_name);
	if (option.nd) {
		char **missing;
		int mnum;
		char *dstd;
		fontset = XCreateFontSet(display, option.fn, &missing, &mnum, &dstd);
		XFreeStringList(missing);
		if (fontset == NULL) {
			fprintf(stderr, "SDClock: theme fontset error..exit.\n");
			exit(EXIT_FAILURE);
		}
	}
	gcm = GCForeground | GCBackground;
	gcv.foreground = hand_pixel;
	gcv.background = back_pixel;
	gc = XCreateGC(display, root, gcm, &gcv);
	XStoreName(display, window, app_name);
	XSetIconName(display, window, app_name);
	XSetCommand(display, window, av, ac);
	XSetWMProtocols(display, main_window, &wm_delete_window, 1);
	XSetWindowBackgroundPixmap(display, main_window, back_pm);
	XShapeCombineMask(display, main_window, ShapeBounding, 0, 0, mask_pm, 
						ShapeSet);
	XMapWindow(display, window);
}

void
UpdateClock(void) {
	int h = old_hour, m = old_minute, s = old_second;
	struct tm *time_struct;
	time_t curtime;
	curtime = time(NULL);
	time_struct = localtime(&curtime);
	h = time_struct->tm_hour;
	m = time_struct->tm_min;
	s = time_struct->tm_sec;
	if (option.ns) {
		if (h == old_hour && m == old_minute && s == old_second)
			return;
	} else {
		if (h == old_hour && m == old_minute)
			return;
	}
	XCopyArea(display, back_pm, all_pm, gc, 0, 0, ix, iy, 0, 0);
	if (option.nd) {
		char date[S_M_LENGTH];
		strftime (date, sizeof(date), option.dl, time_struct);
		XSetForeground(display, gc, date_pixel);
		XmbDrawString(display, all_pm, fontset, gc, option.xd, option.yd,
					date, strlen(date));
	}
	if (option.nn) {
		double angleh, anglem, angles;
		int hx, hy, mx, my, sxx, syy, ex, ey, ehx, ehy, emx, emy;
		if (option.ns && !option.sxy) {
			angles = (M_PI / 30.0) * (double) s;
			sxx = option.sx + (int) ((double) option.sl * sin(angles));
			syy = option.sy - (int) ((double) option.sl * cos(angles));
			XSetForeground(display, gc, second_hand_pixel);
			XSetLineAttributes(display, gc, option.sw, LineSolid, CapRound,
								JoinRound);
			if (option.ne) {
				ex = option.sx - (int) ((double) option.se * sin(angles));
				ey = option.sy + (int) ((double) option.se * cos(angles));
				XDrawLine(display, all_pm, gc, ex, ey, sxx, syy);
			} else {
				XDrawLine(display, all_pm, gc, option.sx, option.sy, sxx, syy);
			}
		}
		angleh = (M_PI / 6.0) * (double) h + (M_PI / 360.0) * (double) m;
		hx = option.xa + (int) ((double) option.hl * sin(angleh));
		hy = option.ya - (int) ((double) option.hl * cos(angleh));
		anglem = (M_PI / 30.0) * (double) m;
		mx = option.xa + (int) ((double) option.ml * sin(anglem));
		my = option.ya - (int) ((double) option.ml * cos(anglem));
		XSetForeground(display, gc, hand_pixel);
		XSetLineAttributes(display, gc, option.hw, LineSolid, CapRound, 
							JoinRound);
		if (option.ne) {
			ehx = option.xa - (int) ((double) option.he * sin(angleh));
			ehy = option.ya + (int) ((double) option.he * cos(angleh));
			emx = option.xa - (int) ((double) option.he * sin(anglem));
			emy = option.ya + (int) ((double) option.he * cos(anglem));
			XDrawLine(display, all_pm, gc, ehx, ehy, hx, hy);
			XDrawLine(display, all_pm, gc, emx, emy, mx, my);
		} else {
			XDrawLine(display, all_pm, gc, option.xa, option.ya, hx, hy);
			XDrawLine(display, all_pm, gc, option.xa, option.ya, mx, my);
		}
		XSetForeground(display, gc, second_hand_pixel);
		if (option.ns && option.sxy) {
			angles = (M_PI / 30.0) * (double) s;
			sxx = option.xa + (int) ((double) option.sl * sin(angles));
			syy = option.ya - (int) ((double) option.sl * cos(angles));
			XSetLineAttributes(display, gc, option.sw, LineSolid, CapRound,
								JoinRound);
			if (option.ne) {
				ex = option.xa - (int) ((double) option.se * sin(angles));
				ey = option.ya + (int) ((double) option.se * cos(angles));
				XDrawLine(display, all_pm, gc, ex, ey, sxx, syy);
			} else {
				XDrawLine(display, all_pm, gc, option.xa, option.ya, sxx, syy);
			}
		}
		XFillArc(display, all_pm, gc, option.xa - (option.ar / 2), 
			option.ya - (option.ar / 2), option.ar, option.ar, 0, 360 * 64);
	}
	XSetWindowBackgroundPixmap(display, main_window, all_pm);
	XClearWindow(display, main_window);
	old_hour = h;
	old_minute = m;
	old_second = s;
}

void
HandleEvents(int *should_quit) {
	XEvent event;
	*should_quit = FALSE;
	FILE *fpx, *fpy;
	const char *filex, *filey;
	filex = "/root/.SDClock/conf/dx";
	filey = "/root/.SDClock/conf/dy";
	while (XPending(display)) {
		XNextEvent(display, &event);
		switch (event.type) {
			case DestroyNotify:
				*should_quit = TRUE;
				break;
			case ClientMessage:
				if (event.xclient.data.l[0] == wm_delete_window)
					*should_quit = TRUE;
				break;
			case ButtonPress:
				if (event.xbutton.button == 1) {
					XRaiseWindow(display, window);
					org_x = event.xbutton.x;
					org_y = event.xbutton.y;
				break;
				case MotionNotify:
					XMoveWindow(display, window, event.xmotion.x_root - org_x,
					event.xmotion.y_root - org_y);
					break;
				}
				if(event.xbutton.button == 2) {
					pid_t pid = fork();
					switch (pid) {
						case 0:
							execlp ("/bin/sh", "sh", "-c", option.mm, NULL);
							fprintf (stderr, "SDClock: M Exec failed.\n");
							exit (0);
							break;
						case -1:
							fprintf (stderr, "SDClock: M Fork failed.\n");
							break;
					}
				}
				if (event.xbutton.button == 3) {
					pid_t pid = fork();
					switch (pid) {
						case 0:
							execlp ("/bin/sh", "sh", "-c", option.rr, NULL);
							fprintf (stderr, "SDClock: R Exec failed.\n");
							exit (0);
							break;
						case -1:
							fprintf (stderr, "SDClock: R Fork failed.\n");
							break;
					}
				}
				break;
			case ButtonRelease:
				if (event.xbutton.button == 1) {
					pos_x = event.xbutton.x_root - event.xbutton.x;
					pos_y = event.xbutton.y_root - event.xbutton.y;
					if ((fpx = fopen(filex, "w")) == NULL) {
						fprintf (stderr, "SDClock: x conf not saved");
						return;
					}
					fprintf (fpx, "%d", pos_x);
					fclose (fpx);
					if ((fpy = fopen(filey, "w")) == NULL) {
						fprintf (stderr, "SDClock: y conf not saved");
						return;
					}
					fprintf (fpy, "%d", pos_y);
					fclose (fpy);
				}
				break;
			default:
				break;
		}
	}
}

static Pixel
GetColor(char *name) {
	XColor color;
	XWindowAttributes attributes;
	XGetWindowAttributes(display, root, &attributes);
	color.pixel = 0;
	if (!XParseColor(display, attributes.colormap, name, &color)) {
		fprintf(stderr, "SDClock: can't parse color ``%s''.\n", name);
	} else if (!XAllocColor(display, attributes.colormap, &color)) {
		fprintf(stderr, "SDClock: can't allocate ``%s''.\n", name);
	}
	return color.pixel;
}

void
DestroyWindow(void) {
	XFreeGC(display, gc);
	if (fontset != NULL) {
		XFreeFontSet(display, fontset);
	}
	XFreePixmap(display, all_pm);
	XFreePixmap(display, back_pm);
	XFreePixmap(display, mask_pm);
	XDestroyWindow(display, window);
	XCloseDisplay(display);
}

